//
//  InfoPanelController.m
//  Perforce
//
//  Created by Adam Czubernat on 21/05/14.
//  Copyright (c) 2014 Perforce Software, Inc. All rights reserved.
//

#import "InfoPanelController.h"
#import "PSDirectorySizeOperation.h"

@interface InfoPanelController () <NSWindowDelegate, NSOutlineViewDataSource, NSOutlineViewDelegate> {
	NSMutableArray *outlineItems;
	__weak IBOutlet NSOutlineView *outlineView;
	NSOperation *sizeOperation;
}
- (NSArray *)generalInfoForItems:(NSArray *)items;
- (NSArray *)metadataForItems:(NSArray *)items;
- (NSArray *)fileInfoForItems:(NSArray *)items;
- (NSDictionary *)detailInfoWithName:(NSString *)name value:(id)value;
- (void)expandItems;
- (void)expandWindowFrame;
- (IBAction)outlineViewExpandAction:(NSButton *)sender;
@end

@implementation InfoPanelController
@synthesize icon;

- (NSString *)windowNibName {
	return NSStringFromClass([self class]);
}

- (void)windowDidLoad {
    [super windowDidLoad];
	[self.window setDelegate:self];
	[(NSPanel *)self.window setBecomesKeyOnlyIfNeeded:YES];
}

- (void)windowWillClose:(NSNotification *)notification {
	[sizeOperation cancel];
	sizeOperation = nil;
}

#pragma mark - Public

- (void)setItems:(NSArray *)items {
	
	// Cancel size calculation
	[sizeOperation cancel];

	[self willChangeValueForKey:@"outlineItems"];
	outlineItems = [NSMutableArray array];
	
	P4Item *firstItem = [items firstObject];
	if (!items.count) {
		[outlineItems addObject:@{ @"cell" : @"NoSelectionCell" }];
	} else {
		[outlineItems addObjectsFromArray:
		 @[
		   @{ @"cell" : @"HeaderCell",
			  @"name" : (items.count == 1 ? firstItem.name :
						 [NSString stringWithFormat:@"%ld documents", items.count]),
			  @"icon" : firstItem.icon },
		   @{ @"cell" : @"GroupCell",
			  @"name" : @"General",
			  @"items": [self generalInfoForItems:items] },
		   @{ @"cell" : @"GroupCell",
			  @"name" : @"Status",
			  @"items": [self metadataForItems:items] },
		   @{ @"cell" : @"GroupCell",
			  @"name" : @"More Info",
			  @"items": [self fileInfoForItems:items] },
		   ]
		 ];
	}
	
	[self didChangeValueForKey:@"outlineItems"];
	[self performSelectorOnMainThread:@selector(expandItems) withObject:nil waitUntilDone:NO];
}

#pragma mark - Private

- (NSArray *)generalInfoForItems:(NSArray *)items {

	NSFileManager *filemanager = [NSFileManager defaultManager];
	P4Item *firstItem = [items firstObject];
	
	NSString *kind, *where;
	NSDate *created, *modified;
	NSMutableDictionary *sizeDetail = [[self detailInfoWithName:@"Size" value:@"Loading..."]
									   mutableCopy];
	
	if (items.count > 1) {
		
		where = [firstItem parent].path;
		kind = [NSString stringWithFormat:@"%ld documents", items.count];

	} else {
				
		where = [firstItem.path stringByDeletingPath];
		NSString *path = firstItem.localPath;
		
		if ([filemanager fileExistsAtPath:path]) {
			MDItemRef itemRef = MDItemCreate(NULL, (CFStringRef)path);
			kind = CFBridgingRelease(MDItemCopyAttribute(itemRef, kMDItemKind));
			created = CFBridgingRelease(MDItemCopyAttribute(itemRef, kMDItemContentCreationDate));
			modified = CFBridgingRelease(MDItemCopyAttribute(itemRef, kMDItemContentModificationDate));
			CFRelease(itemRef);
		} else {
			kind = firstItem.isDirectory ? @"Folder" : @"Untracked file";
		}
	}
	
	NSArray *paths = [items valueForKey:@"path"];
	
	if ([[firstItem path] hasPrefix:@"//"]) {
		// Depot files
		sizeOperation = [[P4Workspace sharedInstance]
						 calculateSizeOfPaths:paths
						 response:^(P4Operation *operation, NSArray *response) {
							 sizeOperation = nil;
							 if (operation.errors)
								 return;
							 
							 NSNumber *count = [response valueForKeyPath:@"@sum.fileCount"];
							 NSNumber *size = [response valueForKeyPath:@"@sum.fileSize"];
							 
							 [self willChangeValueForKey:@"outlineItems"];
							 NSString *total = [NSString stringWithFormat:@"%@ for %ld items",
												[NSString stringWithByteCount:size.integerValue],
												count.integerValue];
							 [sizeDetail setObject:total forKey:@"value"];
							 [self didChangeValueForKey:@"outlineItems"];
						 }];
	} else {
		// Local files
		sizeOperation = [PSDirectorySizeOperation
						 operationWithPaths:paths
						 block:^(long long size, NSInteger count) {
							 sizeOperation = nil;
							 [self willChangeValueForKey:@"outlineItems"];
							 NSString *total = [NSString stringWithFormat:@"%@ for %ld items",
												[NSString stringWithByteCount:size], count];
							 [sizeDetail setObject:total forKey:@"value"];
							 [self didChangeValueForKey:@"outlineItems"];
						 }];
		
		NSOperationQueue *queue = [[NSOperationQueue alloc] init];
		[queue addOperation:sizeOperation];
	}
	
	return [NSArray arrayWithObjects:
			[self detailInfoWithName:@"Kind" value:kind],
			sizeDetail,
			[self detailInfoWithName:@"Where" value:where],
			!created ? nil : // Break an array here if there's no additional data
			[self detailInfoWithName:@"Created" value:created],
			[self detailInfoWithName:@"Modified" value:modified],
			nil];
}

- (NSArray *)metadataForItems:(NSArray *)items {
	P4Item *item = [items firstObject];
	NSMutableDictionary *metadata = item.metadata.mutableCopy;

	// Don't show metadata for multiple files or if file isn't tracked
	if (items.count > 1 || metadata.count == 0)
		return @[ @{ @"cell" : @"EmptyCell" } ];
	
	// Transform values
	NSArray *dateKeys = @[ @"headModTime" ];
	for (NSString *key in dateKeys) {
		id value = [metadata objectForKey:key];
		if (value) {
			value = [NSDate dateWithTimeIntervalSince1970:[value doubleValue]];
			[metadata setObject:value forKey:key];
		}
	}
	NSArray *byteKeys = @[ @"fileSize" ];
	for (NSString *key in byteKeys) {
		id value = [metadata objectForKey:key];
		if (value) {
			value = [NSString stringWithByteCount:[value integerValue]];
			[metadata setObject:value forKey:key];
		}
	}
	
	// Supported keys
	NSArray *keys = @[ @"action", @"headRev",	@"headChange", @"otherOpen0",	@"otherAction0", @"depotFile",	@"dir",			@"fileSize",	@"headModTime",  @"attr-tags" ];
	NSArray *names = @[@"Action", @"Revision",	@"Changelist", @"Opened By",	@"Action",		 @"Depot Path",	@"Depot Path",	@"Depot Size",	@"Modified In",  @"Tags" ];
	NSMutableArray *infoItems = [NSMutableArray array];
	[keys enumerateObjectsUsingBlock:^(id key, NSUInteger idx, BOOL *stop) {
		id name = [names objectAtIndex:idx];
		id value = [metadata objectForKey:key];
		if (value)
			[infoItems addObject:[self detailInfoWithName:name ?: key value:value]];
	}];
	
//	NSMutableDictionary *unused = metadata.mutableCopy;
//	[unused removeObjectsForKeys:keys];
//	PSLog(@"Unused keys %@", unused);
	
	return infoItems;
}

- (NSArray *)fileInfoForItems:(NSArray *)items {
	
	P4Item *item = [items firstObject];
	NSString *path;
	MDItemRef itemRef;
	
	// Don't show info for multiple files or if file isn't tracked
	if (items.count > 1 ||
		!(path = item.localPath) ||
		!(itemRef = MDItemCreate(NULL, (CFStringRef)path))) {
		return @[ @{ @"cell" : @"EmptyCell" } ];
	}

	NSDictionary *fileMetadata;
	CFArrayRef arrayRef = MDItemCopyAttributeNames(itemRef);
	fileMetadata = CFBridgingRelease(MDItemCopyAttributes(itemRef, arrayRef));
	CFRelease(arrayRef);
	CFRelease(itemRef);
	
	// Supported keys
	NSArray *keys =
	@[ @"kMDItemDateAdded",
	   @"kMDItemPixelWidth", @"kMDItemPixelHeight", @"kMDItemResolutionHeightDPI", @"kMDItemHasAlphaChannel", @"kMDItemProfileName", @"kMDItemLayerNames",
	   @"kMDItemTitle", @"kMDItemNumberOfPages", @"kMDItemPageWidth", @"kMDItemPageHeight",
	   @"kMDItemCreator", ];
	NSMutableArray *infoItems = [NSMutableArray array];
	[keys enumerateObjectsUsingBlock:^(id key, NSUInteger idx, BOOL *stop) {
		id value = [fileMetadata objectForKey:key];
		id name = CFBridgingRelease(MDSchemaCopyDisplayNameForAttribute((CFStringRef)key));
		if (name && value)
			[infoItems addObject:[self detailInfoWithName:name value:value]];
	}];
	
//	NSMutableDictionary *unused = fileMetadata.mutableCopy;
//	[unused removeObjectsForKeys:keys];
//	PSLog(@"Unused keys %@", unused);

	return infoItems;
}

- (NSDictionary *)detailInfoWithName:(NSString *)name value:(id)value {
	
	NSString *identifier = @"DetailCell";
	if ([value isKindOfClass:[NSDate class]]) {
		identifier = @"DetailDateCell";
	} else if ([value isKindOfClass:[NSArray class]]) {
		value = [value componentsJoinedByString:@", "];
	} else if ([value isKindOfClass:[NSNumber class]]) {
		value = [value description];
	}
	
	return @{ @"cell" : identifier,
			  @"name" : name,
			  @"value" : value ?: @"null", };
}

#pragma mark OutlineView Expanding

- (void)expandItems {
	[outlineView expandItem:nil expandChildren:YES];
	[self expandWindowFrame];
}

- (void)expandWindowFrame {
	// Get content rect
	NSInteger lastRow = [outlineView numberOfRows]-1;
	CGRect contentRect = CGRectUnion(CGRectZero, [outlineView rectOfRow:lastRow]);
	contentRect = [self.window frameRectForContentRect:contentRect];
	CGFloat height = fmin(contentRect.size.height, [self.window maxSize].height);
				
	// Set window frame
	CGRect frame = [self.window frame];
	frame.origin.y += frame.size.height - height;
	frame.size.height = height;
	[[self.window animator] setFrame:frame display:YES];
}

- (IBAction)outlineViewExpandAction:(NSButton *)sender {
	NSInteger row = [outlineView rowForView:sender];
	id item = [outlineView itemAtRow:row];
	if ([outlineView isItemExpanded:item])
		[[outlineView animator] collapseItem:item];
	else
		[[outlineView animator] expandItem:item];
	[self expandWindowFrame];
}

#pragma mark - NSOutlineView delegate

- (NSView *)outlineView:(NSOutlineView *)view viewForTableColumn:(NSTableColumn *)tableColumn item:(id)item {
	NSDictionary *dictionary = [item representedObject];
	NSString *identifier = [dictionary valueForKey:@"cell"];
	NSTableCellView *cell = [outlineView makeViewWithIdentifier:identifier owner:self];
	return cell;
}

- (CGFloat)outlineView:(NSOutlineView *)view heightOfRowByItem:(id)item {
	NSDictionary *dictionary = [item representedObject];
	NSString *identifier = [dictionary objectForKey:@"cell"];
	return [outlineView rowHeightForIdentifier:identifier];
}

@end
